module Saper
  class Action
    
    INPUT_TYPES = %i(text atom document html json markdown nothing time url xml)
    
    # Tracks subclasses of Saper::Action.
    # @return [Class]
    def self.inherited(base)
      subclasses[base.type] = base
    end
    
    # Returns a hash of subclasses.
    # @return [Hash]
    def self.subclasses
      @subclasses ||= {}
    end
    
    # Returns class name as an underscored string.
    # @return [String]
    def self.type
      name.split("::").last.gsub(/([a-z])([A-Z])/,'\1_\2').downcase
    end
    
    # Returns a subclass with specified type.
    # @param type [Symbol] action type
    # @return [Saper::Action]
    def self.[](type)
      subclasses[type.to_s] || raise(ActionNotFound, type)
    end
    
    # Defines a new argument.
    # @param type [Symbol] argument type (e.g. :text, :xpath)
    # @param options [Hash] argument options (e.g. :optional => true)
    # @return [void]
    def self.argument(type, options = {})
      if Argument.exists?(type)
        arguments.push options.merge(:type => type)
      else
        raise(InvalidType, "Invalid action argument: %s" % type)
      end
    end
    
    # Returns a list of arguments assigned to this action.
    # @return [Array<Hash>]
    def self.arguments
      @arguments ||= []
    end
    
    # Sets (one) or returns (all) acceptable input types.
    # @param input [Symbol]
    # @return [Array<Symbol>]
    def self.accepts(input = nil, options = {})
      if input.nil?
        return @types.nil? ? [] : @types.keys
      end
      if input == :anything
        INPUT_TYPES.each { |type| accepts(type, options) }
      end
      output = options[:returns] || input
      @types ||= {}
      @types[input] = output
    end
    
    # Returns `true` if action accepts specified type as input.
    # @param type [Symbol]
    # @return [Boolean]
    def self.accepts?(type)
      accepts.include?(type)
    end
    
    # Saves Proc that encapsulates action logic and will be used later for data processing.
    # @return [void]
    def self.run(&block)
      @block ||= block
    end
    
    # Returns the type of output for specfied input type.
    # @param type [Symbol] argument type (e.g. :text, :xpath)
    # @return [Array<Symbol>]
    def self.output_for(type)
      accepts?(type) ? @types[type] : :nothing
    end
    
    # Returns `true` if action returns multiple items. Note that this method will report incorrect data for some actions. Use #multiple? instead.
    def self.returns_multiple_items?
      @multiple == true
    end
    
    # Sets a flag, indicating that this action returns multiple items.
    # @return [void]
    def self.returns_multiple_items!
      @multiple = true
    end
    
    # Returns a new instance of Saper::Action.
    # @param data [Hash, Array<Hash>]
    # @return [Saper::Action]
    def self.unserialize(data, namespace = nil, &block)
      if data.is_a?(Array)
        return data.map { |item| unserialize(item, namespace) }
      end
      unless data.is_a?(Hash)
        raise InvalidAction.new(data)
      end
      new(data[:type], *data[:args], :namespace => namespace) do |action|
        if block_given?
          yield action
        end
      end
    end
    
    # Returns a new instance of Saper::Action.
    # @return [Saper::Action]
    def self.new(*args, &block)
      if self == Action
        self[args.shift].new(*args, &block)
      else
        super(*args, &block)
      end
    end
    
    # @todo
    attr_reader :options
    
    # @todo
    attr_reader :arguments
    
    # Returns a new instance of Saper::Action.
    # @return [Saper::Action]
    def initialize(*args)
      @arguments = []
      @options   = {}
      if args.last.is_a?(Hash)
        @options = args.pop
      end
      self.class.arguments.each_with_index do |opts, i|
        opts.merge!(:value => args[i], :action => self)
        @arguments << Argument.new(opts[:type], opts)
      end
      if block_given?
        yield self
      end
    end
    
    # Runs action and returns results.
    # @param input [object] input
    # @return [void] depends on action type
    def run(input = nil, runtime = nil)
      runtime ||= Runtime.new
      input  = validate_input(input)
      output = block.call(runtime, input, *args)
      validate_output output, output_for(input.type)
    end
    
    # Returns human readable action name.
    # @return [String]
    def name
      self.class.name.split("::").last.gsub(/([a-z])([A-Z])/,'\1 \2')
    end
    
    # Returns Saper::Namespace instance.
    # @return [Namespace]
    def namespace
      @options[:namespace].is_a?(Namespace) ? @options[:namespace] : nil
    end
    
    # Returns `true` if action accepts specified type as input.
    # @param type [Symbol]
    # @return [Boolean]
    def accepts?(type)
      self.class.accepts.include?(type)
    end
    
    # Returns the type of output for specfied input type.
    # @param type [Symbol] argument type (e.g. :text, :xpath)
    # @return [Array<Symbol>]
    def output_for(type)
      out = self.class.output_for(type); out.is_a?(Proc) ? out.call(self) : out
    end
    
    # Returns values of action arguments.
    # @return [Array]
    def args
      @arguments.map(&:value)
    end
    
    # Returns a serialized representation of this action.
    # @return [Hash]
    def serialize
      { :type => self.class.type, :args => @arguments.map(&:serialize) }
    end
    
    # Returns a JSON representation of this action.
    # @return [String]
    def to_json(*args)
      serialize.to_json(*args)
    end
    
    # Returns `true` if action returns multiple items.
    # @return [Boolean]
    def multiple?
      self.class.returns_multiple_items?
    end
    
    # Returns Proc that encapsulates action logic (i.e. processes data).
    # @return [Proc]
    def block
      self.class.run || Proc.new { |runtime, input, *args| input }
    end
    
    # Returns string representation of Action.
    # @return [String]
    def to_string
      "\t%s %s" % [self.class.type, @arguments.map(&:to_string).join(", ")]
    end
    
    private
    
    # @todo
    def validate_output(output, type)
      if output.is_a?(Array) && output.empty?
        raise InvalidOutput
      end
      if output.is_a?(Array)
        return output.map { |o| validate_output(o, type) }
      end
      if output.is_a?(Items::Nothing) && output.filtered?
        return output
      end
      unless output.is_a?(Item)
        output = Item.try(type, output)
      end
      if output.is_a?(Item) && output.type == type
        return output
      end
      raise InvalidOutput
    end
    
    # @todo
    def validate_input(input)
      unless input.is_a?(Item)
        input = self.class.accepts.map { |type| Item.try(type, input) }.compact.first
      end
      if input.nil? && accepts?(:nothing)
        return Items::Nothing.new
      end
      if input.nil?
        raise(InvalidInput, input)
      end
      unless accepts?(input.type)
        raise(InvalidInput, input)
      end
      input
    end
    
  end
end
