module Saper
  class Argument
    
    # Tracks subclasses of Saper::Argument.
    # @return [Class]
    def self.inherited(base)
      subclasses[base.type] = base
    end
    
    # Returns a hash of subclasses.
    # @return [Hash]
    def self.subclasses
      @subclasses ||= {}
    end
    
    # Returns class name as an underscored string.
    # @return [String]
    def self.type
      name.split("::").last.gsub(/([a-z])([A-Z])/,'\1_\2').downcase
    end
    
    # Returns a subclass with specified type.
    # @param type [Symbol] action type
    # @return [Saper::Argument]
    def self.[](type)
      subclasses[type.to_s] || raise(InvalidType, "Invalid argument: %s" % type)
    end
    
    # Returns `true` if there is a subclass with specified type.
    # @param type [Symbol] action type
    # @return [Boolean]
    def self.exists?(type)
      subclasses.keys.include?(type.to_s)
    end
    
    # Returns a new instance of Saper::Argument.
    # @return [Saper::Argument]
    def self.new(*args, &block)
      if self == Argument
        self[args.shift].new(*args, &block)
      else
        super(*args, &block)
      end
    end
    
    # Returns a new instacne of Saper::Argument.
    # @option opts[Object] :value
    # @option opts[Saper::Action] :action
    # @option opts[Boolean] :optional
    # @return [Saper::Argument]
    def initialize(opts = {})
      @value = nil
      @opts  = opts
      if mandatory? || !@opts[:value].nil?
        set @opts.delete(:value)
      end
    end
    
    # Sets new value and returns self.
    # @param value [Object]
    # @return [self]
    def set(value)
      if value.is_a?(Saper::Argument)
        value = value.value
      end
      unless valid?(value)
        raise InvalidArgument, value
      else
        @value = normalize(value)
      end
      self
    end
    
    # Returns true if supplied value is accepted by this argument type. Subclasses should override this method.
    # @param value [Object]
    # @return [Boolean]
    def valid?(value)
      true
    end
    
    # Returns normalized value.
    # Subclasses should overrid this method. Subclasses may override this method.
    # @param value [Object]
    # @param value [Object]
    def normalize(value)
      value
    end
    
    # Returns argument value. Subclasses may override this method.
    # @return [Object]
    def value
      @value
    end
    
    # Returns serialized representation of argument's value. Subclasses may override this method.
    def serialize
      value
    end
    
    # Returns true if argument value is mandatory.
    # @return [Boolean]
    def mandatory?
      not optional?
    end
    
    # Returns true if argument value is optional.
    # @return [Boolean]
    def optional?
      @opts[:optional] == true
    end
    
    # Returns parent Action instance that this Argument belongs to.
    # @return [Saper::Action]
    def action
      @opts[:action]
    end
    
    # Returns String representation of argument's value.
    # @return [String]
    def to_string
      value.to_s.inspect
    end
    
  end
end
