module Saper
  class Browser
  
    # Array of requested URLs.
    attr_reader :history
    
    # Approximate number of bytes received.
    attr_reader :received
    
    # Approximate number of bytes sent.
    attr_reader :sent
    
    AGENTS = {
      :ie6     => 'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)',
      :ie7     => 'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)',
      :ie8     => 'Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 1.1.4322; .NET CLR 2.0.50727)',
      :ie9     => 'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)',
      :mozilla => 'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.4b) Gecko/20030516 Mozilla Firebird/0.6',
      :safari  => 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_2) AppleWebKit/534.51.22 (KHTML, like Gecko) Version/5.1.1 Safari/534.51.22',
      :iphone  => 'Mozilla/5.0 (iPhone; U; CPU like Mac OS X; en) AppleWebKit/420+ (KHTML, like Gecko) Version/3.0 Mobile/1C28 Safari/419.3',
      :ipad    => 'Mozilla/5.0 (iPad; U; CPU OS 3_2 like Mac OS X; en-us) AppleWebKit/531.21.10 (KHTML, like Gecko) Version/4.0.4 Mobile/7B334b Safari/531.21.10',
      :android => 'Mozilla/5.0 (Linux; U; Android 3.0; en-us; Xoom Build/HRI39) AppleWebKit/534.13 (KHTML, like Gecko) Version/4.0 Safari/534.13',
      :saper   => 'Mozilla/5.0 (compatible; Saper Ruby client %s)' % Saper::VERSION
    }
    
    # Returns a new Browser instance.
    # @option options [Symbol] :agent User agent
    # @option options [Hash] :headers Additional request headers
    # @option options [Logger] :logger Logger instance
    # @return [Saper::Browser]
    def initialize(options = {})
      @agent    = options.delete(:agent) || :saper
      @headers  = options.delete(:headers)
      @logger   = options.delete(:logger) || Saper::Logger.new
      @history  = []
      @received = 0
      @sent     = 0
      @mech = Mechanize.new do |a|
        a.robots                 = false
        a.user_agent             = agent
        a.request_headers        = headers
        a.pluggable_parser.csv   = nil
        a.pluggable_parser.html  = nil
        a.pluggable_parser.xhtml = nil
        a.pluggable_parser.xml   = nil
      end
      @mech.pre_connect_hook do |agent, req|
        @sent += req.to_hash.to_s.size
        @sent += (req.body.nil? ? 0 : req.body.size)
      end
      @mech.post_connect_hook do |agent, uri, resp, body|
        @received += resp.to_hash.to_s.size
        @received += body.size
      end
      @logger.new_browser(self)
    end
    
    # Returns the number of HTTP requests.
    # @return [Integer]
    def requests
      @history.size
    end
    
    # Returns additional request headers.
    # @return [Hash]
    def headers
      @headers.respond_to?(:to_hash) ? @headers : {}
    end
    
    # Performs a GET request and returns Saper::Document.
    # @param url [String] URL to request
    # @param query [Hash] query options
    # @return [Saper::Document]
    def get(url, query = {})
      @logger.new_get_request(url)
      @history.push url
      data = @mech.get(url, query)
      Saper::Items::Document.new data.body, data.uri, data.header
    rescue Mechanize::ResponseCodeError
      Saper::Items::Nothing.new # TODO: change to custom exception
    end
    
    # Performs a POST request and returns Saper::Document.
    # @param url [String] URL to request
    # @param query [Hash] payload
    # @return [Saper::Document]
    def post(url, query = {}, headers = {})
      @logger.new_post_request(url)
      @history.push url
      data = @mech.post(url, query, headers)
      Saper::Items::Document.new data.body, data.uri, data.header
    rescue Mechanize::ResponseCodeError
      Saper::Items::Nothing.new # TODO: change to custom exception
    end
    
    # @todo
    def post_with_bearer_token(url, query, token)
      post(url, query, { "Authorization" => "Bearer %s" % token })
    end
    
    # Returns User-Agent string used with requests.
    # @return [String]
    def agent
      AGENTS[@agent.to_sym] || @agent.to_s
    end
    
  end
end
