module Saper
  class Item
    
    # Tracks subclasses of Saper::Item.
    # @return [Class]
    def self.inherited(base)
      subclasses[base.type] = base
    end
    
    # Returns a hash of subclasses.
    # @return [Hash]
    def self.subclasses
      @subclasses ||= {}
    end
    
    # Returns class name as an underscored string.
    # @return [String]
    def self.type
      name.split("::").last.gsub(/([a-z])([A-Z])/,'\1_\2').downcase
    end
    
    # Returns a subclass with specified type.
    # @param type [Symbol] action type
    # @return [Saper::Argument]
    def self.[](type)
      subclasses[type.to_s] || raise(InvalidItem, "Invalid action argument: %s" % type)
    end
    
    # Returns `true` if there is a subclass with specified type.
    # @param type [Symbol] action type
    # @return [Boolean]
    def self.exists?(type)
      subclasses.keys.include?(type.to_s)
    end
    
    # Returns a new instance of Saper::Argument. Returns nil and fails
    # silently if there is an error during initialization.
    # @return [Saper::Argument]
    def self.try(*args, &block)
      begin
        new(*args, &block)
      rescue InvalidItem, ArgumentError
        nil
      end
    end
    
    # Returns a new instance of Saper::Argument.
    # @return [Saper::Argument]
    def self.new(*args, &block)
      if self == Item
        self[args.shift].new(*args, &block)
      else
        super(*args, &block)
      end
    end
    
    # Returns `true` if other item contains the same data.
    # @return [Boolean]
    def ==(other)
      to_s == other.to_s
    end
    
    # Returns class name as a Symbol.
    # @return [Symbol]
    def type
      self.class.type.to_sym
    end
    
    # @todo
    def serialize
      to_native
    end
    
  end
end