module Saper
  
  # Runs a recipe saved in a file.
  # @param file [String] file name
  # @param recipe [Symbol] recipe name
  # @param input [Object] input for recipe
  # @return [Saper::Runtime]
  def self.run_from_file(file, recipe, input, opts = {})
    load(file).run(recipe, input, opts)
  end
  
  # Parses recipes saved in a file.
  # @param file [String] file name
  # @return [Saper::Namespace]
  def self.load(path)
    Namespace.parse File.read(path)
  end
  
  # Namespace is a parsing / storage utility for recipes
  # that were meant to work as a group. It must be initiated
  # for some actions to work properly (e.g. `run_chain`).
  class Namespace
    
    # Parses namespace specified as a string or a block.
    # @param string [String] list of recipes
    # @return [Saper::Namespace]
    def self.parse(string = nil, &block)
      instance = DSL.new
      if string.is_a?(String)
        instance.instance_eval(string)
      end
      if block_given?
        instance.instance_eval(&block)
      end
      instance.namespace
    end
    
    # Returns a new instance of Saper::Namespace
    # @param data [Hash]
    # @return [Saper::Namespace]
    def self.unserialize(data, &block)
      unless data.is_a?(Hash)
        raise(InvalidNamespace, data)
      end
      unless data[:recipes].is_a?(Array)
        raise(InvalidNamespace, data)
      end
      new do |namespace|
        data[:recipes].each do |recipe|
          namespace << Recipe.unserialize(recipe, namespace)
        end
        if block_given?
          yield namespace
        end
      end
    end
    
    # Returns a new Namespace instance
    # @return [Saper::Namespace]
    def initialize
      @recipes = {}
      if block_given?
        yield self
      end
    end
    
    # Runs an recipe and returns resulting Saper:Runtime.
    # @param input [Object]
    # @param name [String, Symbol] action name
    # @param options [Hash] options for Runtime instance
    # @return [Saper::Runtime]
    def run(name = nil, input = nil, options ={})
      self[name].run(input, options)
    end
    
    # Returns recipe with specified ID or fails silently if not found.
    # @param id [String, Symbol] recipe ID
    # @return [Saper::Recipe, nil]
    def try(id)
      begin
        self[id]
      rescue
        nil
      end
    end
    
    # Returns recipe with specified ID.
    # @param id [String, Symbol] recipe ID
    # @return [Saper::Recipe]
    def [](id)
      id = id.to_s
      if @recipes.key?(id)
        value = @recipes[id]
      else
        value = find(id)
      end
      if value.is_a?(Recipe)
        return value
      end
      if value.respond_to?(:to_recipe)
        return @recipes[id] = value.to_recipe
      end
      raise RecipeNotFound, id
    end
    
    # Caches recipe instance.
    # @param id [String, Symbol] recipe ID
    # @param recipe [Saper::Recipe]
    # @return [Saper::Recipe]
    def []=(id, recipe)
      unless recipe.is_a?(Saper::Recipe)
        unless recipe.respond_to?(:to_recipe)
          raise ActionExpected
        end
      end
      @recipes[id.to_s] = recipe
    end
    
    # Searches for and returns a recipe.
    # @return [Saper::Recipe]
    def find(id)
      raise RecipeNotFound, id
    end
    
    # Returns the number of recipes within this namespace.
    # @return [Integer]
    def size
      @recipes.size
    end
    
    # Returns a serialized representation of this Namespace.
    # @return [Hash]
    def serialize
      { :recipes => @recipes.values.map(&:serialize) }
    end
    
    # Returns a JSON representation of this recipe.
    # @return [String]
    def to_json(*args)
      serialize.to_json(*args)
    end
    
  end
end
