module Saper
  class Recipe
    
    # Returns a new instance of Saper::Recipe
    # @param data [Hash]
    # @return [Saper::Recipe]
    def self.unserialize(data, namespace = nil, &block)
      if data.is_a?(Array)
        return data.map { |item| unserialize(item, namespace) }
      end
      unless data.is_a?(Hash)
        raise(InvalidRecipe, data)
      end
      new(data[:id], :name => data[:name]) do |recipe|
        unless data[:actions].nil?
          recipe.push *Action.unserialize(data[:actions], namespace)
        end
        if block_given?
          yield recipe
        end
      end
    end
    
    # Parses block and returns a Recipe instance.
    # @param id [Symbol] recipe ID
    # @return [Saper::Recipe]
    def self.parse(id = nil, name = nil, &block)
      DSL::Recipe.parse(id, name, &block)
    end
    
    # Returns unique ID of the recipe.
    attr_reader :id

    # Returns list of recipe actions.
    attr_reader :actions
    
    # Returns a new instance of Saper::Recipe
    # @param id [String]
    # @return [Saper::Recipe]
    def initialize(id = nil, options = {})
      @id      = id || SecureRandom.hex
      @actions = []
      @options = options
      if block_given?
        yield self
      end
    end
    
    # Returns recipe name.
    # @return [String, nil]
    def name
      @options[:name].nil? ? nil : @options[:name].to_s
    end
    
    # Returns `true` if recipe accepts specified type as input.
    # @param type [Symbol]
    # @return [Boolean]
    def accepts?(type)
      actions.each do |action|
        unless action.accepts?(type)
          return false
        end
        type = action.output_for(type)
      end
      true
    end
    
    # Sets (one) or returns (all) acceptable input types.
    # @param input [Symbol]
    # @return [Array<Symbol>]
    def accepts
      Action::INPUT_TYPES.select { |type| accepts?(type) }
    end
    
    # Returns a list of data types required as input.
    # @return [Array<Symbol>]
    def requires
      accepts.include?(:nothing) ? [] : accepts
    end
    
    # Returns `true` if recipe requires some input.
    # @return [Boolean]
    def requires?
      !requires.empty?
    end
    
    # Adds a new action to the end of the recipe and returns self.
    # @param action [Saper::Action]
    # @return [self]
    def <<(action)
      if action.is_a?(Action)
        @actions.push(action)
      else
        raise ActionExpected, action
      end
      self
    end
    
    # Adds new actions to the end of the recipe and returns self.
    # @param actions [Array<Saper::Action>]
    # @return [self]
    def push(*actions)
      actions.compact.each do |action|
        self << action
      end
      self
    end
    
    # Runs recipe and returns Result instance.
    # @param input [object] input for the first action
    # @param runtime [Runtime, Hash] Runtime instance or options for a new one 
    # @return [Saper::Result]
    def run(input = nil, runtime = nil)
      unless runtime.is_a?(Runtime)
        runtime = Runtime.new({}, runtime || {})
      end
      runtime.logger.executing_recipe(self)
      stack = Stack.new(runtime, input)
      enum  = enum_for(:each_stack, stack)
      Result.new(enum)
    end
    
    # Returns `true` if recipe contains no actions.
    # @return [Boolean]
    def empty?
      actions.empty?
    end
    
    # Returns `true` if recipe produces multiple results.
    # @return [Boolean]
    def multiple?
      empty? ? false : actions.any?(&:multiple?)
    end
    
    # Returns a serialized representation of this Recipe.
    # @return [Hash]
    def serialize
      { :id => id, :name => name, :actions => @actions.map(&:serialize) }
    end
    
    # Returns a string representation of this recipe.
    # @return [String]
    def to_string
      "recipe %s do\n%s\nend" % [id.inspect, actions.map(&:to_string).join("\n")]
    end
    
    # Returns a JSON representation of this recipe.
    # @return [String]
    def to_json(*args)
      serialize.to_json(*args)
    end
    
    private
    
    # Yields resulting Stack instances.
    # @param stack [Saper::Stack]
    # @param index [Integer] action undex
    # @yield [Saper::Stack]
    # @return [nil]
    def each_stack(stack, index = 0, &block)
      if stack.error?
        yield stack
        return nil
      end
      if stack.filtered?
        yield stack
        return nil
      end
      if index >= actions.size
        yield stack
        return nil
      end
      each_action_result(stack, actions[index]) do |stack|
        each_stack(stack, index + 1, &block)
      end
      nil
    end
    
    # Yields action results one-by-one.
    # @param stack [Saper::Stack]
    # @param action [Saper::Action]
    # @yield [Saper::Item]
    # @return [nil]
    def each_action_result(stack, action)
      begin
        results = action.run(stack.last, stack.runtime)
      rescue Saper::Error => error
        results = error
      end
      if results.is_a?(Items::Nothing) && results.filtered?
        yield stack.add(results); return nil
      end
      unless results.is_a?(Array)
        yield stack.add(results); return nil
      end
      if results.empty?
        yield stack.add(Items::Nothing.new)
      end
      results.each do |result|
        yield stack.copy.add(result)
      end
      nil
    end
    
  end
end