module Saper
  # Runtime provides a set of auxiliary functions to Saper actions:
  # browser requests, keychain access, storage of variables.
  class Runtime
    
    # Returns Saper::Browser instance used by runtime.
    # @return [Saper::Browser]
    attr_reader :browser
    
    # Returns Saper::Keychain instance used by runtime.
    # @return [Saper::Keychain]
    attr_reader :keychain
    
    # Returns a Hash of variables stored by runtime.
    # @return [Hash]
    attr_reader :variables
    
    # @todo
    attr_reader :logger
    
    # Returns a new Saper::Runtime instance.
    # @param variables [Hash]
    # @param options [Hash]
    # @option options [Browser] :browser Browser instance
    # @option options [String, Keychain] :keychain Keychain file or instance
    # @option options [String, Symbol] :agent User-Agent type or string
    # @return [Saper::Runtime]
    def initialize(variables = {}, options = {})
      unless options.is_a?(Hash)
        options = {}
      end
      if options[:logger].is_a?(Logger)
        @logger = options[:logger]
      else
        @logger = Logger.new
      end
      if options[:browser].is_a?(Browser)
        @browser = options[:browser]
      else
        @browser = Browser.new(:agent => options[:agent], :logger => logger)
      end
      if options[:keychain].is_a?(Keychain)
        @keychain = options[:keychain]
      end
      if options[:keychain].is_a?(String)
        @keychain = Keychain.load(options[:keychain], :logger => logger)
      end
      unless @keychain.is_a?(Keychain)
        @keychain = Keychain.new({}, :logger => logger)
      end
      @variables = variables
      if block_given?
        yield self
      end
    end
    
    # Returns volume of incoming traffic (in bytes)
    # @return [Integer]
    def bytes_received
      browser.received
    end
    
    # Returns volume of outgoing traffic (in bytes)
    # @return [Integer]
    def bytes_sent
      browser.sent
    end
    
    # Returns number of HTTP requests
    # @return [Integer]
    def http_requests
      browser.requests
    end
    
    # Get variable value
    # @param name [Symbol]
    # @return [Object]
    def [](name)
      @variables[name]
    end
    
    # Set variable value
    # @param name [Symbol]
    # @param value [Object]
    # @return [Object]
    def []=(name, value)
      @variables[name] = value
    end
    
    # Returns access credentials for specified service
    # @param service [Symbol]
    # @return [Object, nil]
    def credentials(service)
      keychain[service]
    end
    
    # Returns a duplicate instance of self.
    # @return [Saper::Runtime]
    def copy
      Runtime.new(variables.dup, :browser => browser, :keychain => keychain)
    end
    
    # Returns a duplicate instance of self without the stored variables.
    # @return [Saper::Runtime]
    def copy_without_variables
      Runtime.new({}, :browser => browser, :keychain => keychain)
    end
    
  end
end